import os

from django.core.exceptions import ImproperlyConfigured


def get_environment_variable(name):
    """Return the environment variable or return exception"""
    try:
        return os.environ[name]
    except KeyError:
        raise ImproperlyConfigured(f'Set the {name} environment variable')



from factory import Faker, Iterator
from factory.django import DjangoModelFactory

from ..constants.network import ACCEPTED_FEE_LIST, MAX_POINT_VALUE, MIN_POINT_VALUE, VERIFY_KEY_LENGTH
from ..models.network_transaction import NetworkTransaction


class NetworkTransactionFactory(DjangoModelFactory):
    amount = Faker('pyint', max_value=MAX_POINT_VALUE, min_value=MIN_POINT_VALUE)
    fee = Iterator(ACCEPTED_FEE_LIST + [''])
    recipient = Faker('text', max_nb_chars=VERIFY_KEY_LENGTH)

    class Meta:
        model = NetworkTransaction



from factory import Faker

from .network_node import NetworkNodeFactory
from ..constants.network import ACCOUNT_FILE_HASH_LENGTH, BLOCK_IDENTIFIER_LENGTH, MAX_POINT_VALUE, MIN_POINT_VALUE
from ..models.network_validator import NetworkValidator


class NetworkValidatorFactory(NetworkNodeFactory):
    daily_confirmation_rate = Faker('pyint', max_value=MAX_POINT_VALUE, min_value=MIN_POINT_VALUE)
    root_account_file = Faker('url')
    root_account_file_hash = Faker('text', max_nb_chars=ACCOUNT_FILE_HASH_LENGTH)
    seed_block_identifier = Faker('text', max_nb_chars=BLOCK_IDENTIFIER_LENGTH)

    class Meta:
        model = NetworkValidator
        abstract = True



import re

from rest_framework import serializers

from thenewboston.constants.network import (
    ACCEPTED_FEE_LIST,
    MAX_POINT_VALUE,
    MEMO_MAX_LENGTH,
    MIN_POINT_VALUE,
    VERIFY_KEY_LENGTH
)
from thenewboston.utils.serializers import validate_keys


class NetworkTransactionSerializer(serializers.Serializer):
    amount = serializers.IntegerField(max_value=MAX_POINT_VALUE, min_value=MIN_POINT_VALUE)
    fee = serializers.ChoiceField(choices=ACCEPTED_FEE_LIST, required=False)
    memo = serializers.CharField(max_length=MEMO_MAX_LENGTH, required=False)
    recipient = serializers.CharField(max_length=VERIFY_KEY_LENGTH, min_length=VERIFY_KEY_LENGTH)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass

    def validate(self, data):
        """Check that there are no additional keys included in the data"""
        validate_keys(self, data)

        return data

    @staticmethod
    def validate_amount(amount):
        """Check that amount is not 0"""
        if amount == 0:
            raise serializers.ValidationError('Tx amount can not be 0 (Tx should be excluded)')

        return amount

    @staticmethod
    def validate_memo(memo):
        """Check that all characters are valid"""
        if not re.match('^[a-zA-Z0-9_ ]*$', memo):
            raise serializers.ValidationError('Invalid characters in memo')

        return memo

    @staticmethod
    def validate_recipient(recipient):
        """Check that recipient is a valid hexadecimal"""
        try:
            int(recipient, 16)
        except ValueError:
            raise serializers.ValidationError('Recipient must be a valid hexadecimal')

        return recipient



from django_redis.pool import ConnectionFactory


class FakeConnectionFactory(ConnectionFactory):
    """Used for faking Redis, as we don't want to implement fake Redis connection pooling."""

    def get_connection(self, params):
        return self.redis_client_cls(**self.redis_client_cls_kwargs)



from rest_framework import serializers

from thenewboston.constants.network import MAX_POINT_VALUE, MIN_POINT_VALUE, PROTOCOL_CHOICES, VERIFY_KEY_LENGTH

"""
The ConfigurationSerializer is used as a base class to ensure that a node is properly configured
- used during the connection process
"""


class ConfigurationSerializer(serializers.Serializer):
    account_number = serializers.CharField(max_length=VERIFY_KEY_LENGTH)
    default_transaction_fee = serializers.IntegerField(max_value=MAX_POINT_VALUE, min_value=MIN_POINT_VALUE)
    ip_address = serializers.IPAddressField(protocol='both')
    node_identifier = serializers.CharField(max_length=VERIFY_KEY_LENGTH)
    node_type = serializers.CharField(max_length=22)
    port = serializers.IntegerField(allow_null=True, max_value=65535, min_value=0, required=False)
    protocol = serializers.ChoiceField(choices=PROTOCOL_CHOICES)
    version = serializers.CharField(max_length=32)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass



from rest_framework import serializers


def validate_keys(serializer_instance, data):
    """Check that there are no additional keys included in the data"""
    invalid_keys = {k for k in data.keys() if k not in serializer_instance.fields.keys()}

    if invalid_keys:
        raise serializers.ValidationError(f'Invalid keys {invalid_keys}')



from nacl.signing import SigningKey, VerifyKey

from thenewboston.accounts.key_files import (
    create_account_and_save_signing_key_file,
    read_signing_key_file,
    write_signing_key_file
)
from thenewboston.accounts.manage import create_account


def test_create_account_and_save_signing_key_file(tmpdir):
    file_path = tmpdir.join('signing_key_file')
    signing_key, account_number = create_account_and_save_signing_key_file(file_path)

    assert isinstance(signing_key, SigningKey)
    assert isinstance(account_number, VerifyKey)
    assert signing_key == read_signing_key_file(file_path)


def test_read_signing_key_file(tmpdir):
    file_path = tmpdir.join('signing_key_file')
    signing_key, account_number = create_account_and_save_signing_key_file(file_path)

    assert signing_key == read_signing_key_file(file_path)


def test_write_signing_key_file(tmpdir):
    file_path = tmpdir.join('signing_key_file')
    signing_key, account_number = create_account()
    write_signing_key_file(signing_key, file_path)

    assert signing_key == read_signing_key_file(file_path)



import requests

from thenewboston.utils.exceptions import NetworkException


def fetch(*, url, headers):
    """Send a GET request and return response as Python object"""
    response = requests.get(url, headers=headers)
    return validate_response(response)


def patch(*, url, body):
    """Send a PATCH request and return response as Python object"""
    response = requests.patch(url, json=body)
    return validate_response(response)


def post(*, url, body):
    """Send a POST request and return response as Python object"""
    response = requests.post(url, json=body)
    return validate_response(response)


def validate_response(response):
    """
    Validate status code

    Return response as Python object
    """
    if response.status_code >= 400:
        err = f'status_code:{response.status_code} - {response.text}'
        raise NetworkException(err)

    return response.json()



# -*- coding: utf-8 -*-
import argparse
import decimal
import math
from pathlib import Path

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator, validate_ipv4_address, validate_ipv6_address


def str_length_validator(length: int = None, min_len: int = None, max_len: int = None):
    """Argparse validator for strings"""
    def inner(value):
        if not isinstance(value, str):
            raise argparse.ArgumentTypeError('Expecting string')
        if length and not len(value) == length:
            raise argparse.ArgumentTypeError('String length should be exactly %s chars' % length)
        if min_len and len(value) < min_len:
            raise argparse.ArgumentTypeError('String length should be greater or equal to %s chars' % min_len)
        if max_len and len(value) > max_len:
            raise argparse.ArgumentTypeError('String length should be less or equal to %s chars' % max_len)
        return value
    return inner


def int_validator(min_val: int = None, max_val: int = None):
    """Argparse validator for integers"""
    def inner(value):
        try:
            value = int(value)
        except ValueError:
            raise argparse.ArgumentTypeError('Value is not a valid integer')
        if min_val and value < min_val:
            raise argparse.ArgumentTypeError('Value can not be less than %s' % min_val)
        if max_val and value > max_val:
            raise argparse.ArgumentTypeError('Value can not be greater than %s' % max_val)
        return value
    return inner


def decimal_validator(min_val: [int, decimal.Decimal] = None, max_val: [int, decimal.Decimal] = None):
    """Argparse validator for decimals"""
    def inner(value):
        try:
            value = decimal.Decimal(value)
        except decimal.InvalidOperation:
            raise argparse.ArgumentTypeError('Value is not a valid decimal')
        if math.isnan(value):
            raise argparse.ArgumentTypeError('NaN is not allowed')
        if math.isinf(value):
            raise argparse.ArgumentTypeError('Infinity is not allowed')
        if min_val and value < min_val:
            raise argparse.ArgumentTypeError(f'Value can not be less than {min_val}')
        if max_val and value > max_val:
            raise argparse.ArgumentTypeError(f'Value can not be greater than {max_val}')
        return value
    return inner


def ipv46_validator():
    """Argparse validator to check whether a string is IP v4 or v6 address"""
    def inner(value):
        try:
            validate_ipv4_address(value)
        except ValidationError:
            try:
                validate_ipv6_address(value)
            except ValidationError:
                raise argparse.ArgumentTypeError('Enter a valid IPv4 or IPv6 address')
        return value
    return inner


def url_validator(suffix: str = None):
    """Argparse validator to check whether a string a proper url"""
    def inner(value):
        try:
            validator = URLValidator(schemes=['http', 'https'])
            validator(value)
        except ValidationError:
            raise argparse.ArgumentTypeError('Invalid URL')
        if suffix and Path(value).suffix != suffix:
            raise argparse.ArgumentTypeError('JSON file required')
        return value
    return inner



from thenewboston.blocks.signatures import generate_signature
from thenewboston.utils.tools import sort_and_encode
from thenewboston.verify_keys.verify_key import encode_verify_key, get_verify_key


def generate_signed_request(*, data, nid_signing_key):
    """Generate and return signed request"""
    node_identifier = get_verify_key(signing_key=nid_signing_key)
    signature = generate_signature(
        message=sort_and_encode(data),
        signing_key=nid_signing_key
    )
    return {
        'message': data,
        'node_identifier': encode_verify_key(verify_key=node_identifier),
        'signature': signature
    }



from nacl.encoding import HexEncoder
from nacl.signing import SigningKey, VerifyKey


def encode_verify_key(*, verify_key):
    """Return the hexadecimal representation of the binary account number data"""
    if not isinstance(verify_key, VerifyKey):
        raise RuntimeError('verify_key must be of type nacl.signing.VerifyKey')

    return verify_key.encode(encoder=HexEncoder).decode('utf-8')


def get_verify_key(*, signing_key):
    """Return the verify key from the signing key"""
    if not isinstance(signing_key, SigningKey):
        raise RuntimeError('signing_key must be of type nacl.signing.SigningKey')

    return signing_key.verify_key



from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT
from rest_framework.test import APIClient


class APIError(Exception):

    def __init__(self, *args, resp, expected):
        """Inits APIError which includes the response and the expected response"""
        super().__init__(*args)
        self.resp = resp
        self.expected = expected


class UserWrapper:

    def __init__(self, user=None):
        """Creates APIClient and authenticates the user"""
        self.user = user
        self.client = APIClient()
        self.client.force_authenticate(user)

    def _method(self, method, url, data):
        return getattr(self.client, method)(url, data, format='json')

    @staticmethod
    def check_code(resp, expected):
        if resp.status_code != expected:
            msg = '{} {}'.format(resp.status_code, resp.content)[:1000]
            raise APIError(msg, resp=resp, expected=expected)

    def delete(self, url, data=None):
        return self._method('delete', url, data or {})

    def delete_json(self, *args, expected=HTTP_204_NO_CONTENT, **kwargs):
        response = self.delete(*args, **kwargs)
        self.check_code(response, expected)
        assert not response.content, response.content

    def get(self, url, data=None):
        return self._method('get', url, data or {})

    def get_json(self, *args, expected=HTTP_200_OK, **kwargs):
        response = self.get(*args, **kwargs)
        self.check_code(response, expected)
        return response.json()

    def patch(self, url, data):
        return self._method('patch', url, data)

    def patch_json(self, *args, expected=HTTP_200_OK, **kwargs):
        response = self.patch(*args, **kwargs)
        self.check_code(response, expected)
        return response.json()

    def post(self, url, data):
        return self._method('post', url, data)

    def post_json(self, *args, expected=HTTP_201_CREATED, **kwargs):
        response = self.post(*args, **kwargs)
        self.check_code(response, expected)
        return response.json()

    def put(self, url, data):
        return self._method('put', url, data)

    def put_json(self, *args, expected=HTTP_200_OK, **kwargs):
        response = self.put(*args, **kwargs)
        self.check_code(response, expected)
        return response.json()



from tests.helpers import random_encoded_account_number

from thenewboston.accounts.manage import create_account
from thenewboston.blocks.block import generate_block
from thenewboston.constants.network import BANK, PRIMARY_VALIDATOR
from thenewboston.serializers.network_block import NetworkBlockSerializer
from thenewboston.verify_keys.verify_key import encode_verify_key


def test_network_block_serializer():
    signing_key, account_number = create_account()
    encoded_account_number = encode_verify_key(verify_key=account_number)

    transactions = [
        {
            'amount': 1,
            'fee': BANK,
            'recipient': random_encoded_account_number(),
        },
        {
            'amount': 1,
            'fee': PRIMARY_VALIDATOR,
            'recipient': random_encoded_account_number(),
        },
        {
            'amount': 5,
            'memo': 'Hello there I am 123 years old',
            'recipient': random_encoded_account_number(),
        }
    ]

    block = generate_block(
        account_number=account_number,
        balance_lock=encoded_account_number,
        signing_key=signing_key,
        transactions=transactions
    )

    serializer = NetworkBlockSerializer(data=block)
    assert serializer.is_valid()



import pytz
from factory import Faker

from .created_modified import CreatedModifiedFactory
from ..models.confirmation_service import ConfirmationService


class ConfirmationServiceFactory(CreatedModifiedFactory):
    end = Faker('date_time', tzinfo=pytz.utc)
    start = Faker('date_time', tzinfo=pytz.utc)

    class Meta:
        model = ConfirmationService



import json
from hashlib import sha3_256 as sha3

from django.core.files.base import ContentFile
from django.core.files.storage import default_storage


def get_file_hash(file):
    """Return hash value of file"""
    h = sha3()

    with default_storage.open(file, 'rb') as file:
        chunk = 0

        while chunk != b'':
            chunk = file.read(1024)
            h.update(chunk)

    return h.hexdigest()


def read_json(file):
    """Read JSON file"""
    try:
        with default_storage.open(file, 'r') as f:
            data = json.load(f)
    except FileNotFoundError:
        data = None

    return data


def write_json(file, data):
    """Write JSON file"""
    try:
        default_storage.delete(file)
    except NotImplementedError:
        pass

    default_storage.save(
        file,
        ContentFile(
            json.dumps(data)
        )
    )



from nacl.signing import SigningKey


def create_account():
    """
    Create a new account

    Return signing_key, account_number
    """
    signing_key = SigningKey.generate()
    account_number = signing_key.verify_key
    return signing_key, account_number



from thenewboston.accounts.manage import create_account
from thenewboston.verify_keys.verify_key import encode_verify_key


def random_encoded_account_number():
    signing_key, account_number = create_account()
    return encode_verify_key(verify_key=account_number)
